{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 0,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 0,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "#### import sys\n",
    "\n",
    "class Alchemist:\n",
    "    def __init__(self, filename, resnum, resname, ffbonded, rtp, charge_file=None, types_file=None):\n",
    "        \"\"\"\n",
    "        alch = Alchemist('file.itp', 10)\n",
    "        \"\"\"\n",
    "        self.filename = filename\n",
    "        self.deprot = {\"ASH\":\"ASP\", \"GLH\":\"GLU\", \"GLU\":\"GLU\", \"ASP\":\"GLU\", \"HSP\":\"HSD\", \"GLUP\":\"GLU\", \"ASPP\":\"ASP\"}\n",
    "        self.resnum = resnum\n",
    "        self.resname = resname\n",
    "        self.rtp = rtp\n",
    "        self.ffbonded = [l for l in open(ffbonded)]\n",
    "        self.content = [line for line in open(self.filename)]\n",
    "        self.sections, self.sections_names = self.get_sections()\n",
    "        self.names_to_types_A, self.nums_to_names, self.names_to_nums = self.get_types()\n",
    "        if charge_file and types_file:\n",
    "            self.charges = {line.split()[0]:line.split()[1] for line in open(charge_file)}\n",
    "            self.names_to_types_B = self.parse_types(types_file)\n",
    "        else:\n",
    "            self.charges, self.names_to_types_B = self.parse_rtp()\n",
    "        self.different_types = self.find_different_types()\n",
    "        self.prefix = 'alch-'\n",
    "    \n",
    "    def __call__(self):\n",
    "        \"\"\"\n",
    "        alch()\n",
    "        \"\"\"\n",
    "        self.add_state_B()\n",
    "        self.all_sections_add_B()\n",
    "        self.write()\n",
    "    \n",
    "    def parse_rtp(self):\n",
    "        # read and return charges & names_to_types\n",
    "        with open(self.rtp, 'r') as f:\n",
    "            lista=[]\n",
    "            x = 0\n",
    "            for line in f:\n",
    "                if len(line.split()) > 1 and line.split()[0] == '[' and line.split()[1] == self.deprot[self.resname]: \n",
    "                    x=1\n",
    "                if x==1 and line.split()[0] != '[':\n",
    "                    lista.append(line)\n",
    "                if len(line.split()) > 1 and line.split()[0] == '[' and line.split()[1] == 'bonds' and x==1:\n",
    "                    x=0\n",
    "                    break\n",
    "        charges_dict={}\n",
    "        name_types_dict={}\n",
    "        for line in lista:\n",
    "            name_types_dict[line.split()[0]]= line.split()[1]\n",
    "            charges_dict[line.split()[0]]=line.split()[2]\n",
    "        return charges_dict, name_types_dict\n",
    "    \n",
    "    def get_types(self):\n",
    "        # return a dict of names:types in state A\n",
    "        # self.names_to_types_A, self.nums_to_names, self.names_to_nums = self.get_types()\n",
    "        section = [l for l in self.get_section_content('atoms') if len(l.split()) > 5 and l.split()[2] == str(self.resnum) and l.split()[3] == self.resname and not l.lstrip().startswith('[') and not l.lstrip().startswith(';')]\n",
    "        ntta = {l.split()[4]:l.split()[1] for l in section}\n",
    "        ntna = {l.split()[0]:l.split()[4] for l in section}\n",
    "        natn = {l.split()[4]:l.split()[0] for l in section}\n",
    "        return ntta, ntna, natn\n",
    "    \n",
    "    def find_different_types(self):\n",
    "        diffs = []\n",
    "        for num in self.nums_to_names.keys():\n",
    "            print(num)\n",
    "            name = self.nums_to_names[num]\n",
    "            try:\n",
    "              if self.names_to_types_A[name] != self.names_to_types_B[name]:\n",
    "                diffs.append(num)\n",
    "            except KeyError:\n",
    "                diffs.append(num)\n",
    "        return diffs\n",
    "    \n",
    "    def parse_types(self, types_file):\n",
    "        #nazwa atomu pierwsza kolumna, typ druga kolumna\n",
    "        types = [line for line in open(types_file)]\n",
    "        return {line.split()[0]:line.split()[1] for line in types if len(line.split())>1}\n",
    "    \n",
    "    def get_sections(self):\n",
    "        # petla po content\n",
    "        # dwie listy\n",
    "        # jesli zaczyna sie od '[', to nazwa do listy z nazwami, a pusta lista do listy z sekcjami\n",
    "        # dodajemy kolejne linijki do ostatniej podlisty\n",
    "        names,section_insides =[],[]\n",
    "        section_insides.append([])\n",
    "        names.append('header')\n",
    "        for line in self.content:\n",
    "            if line.strip().startswith('['):\n",
    "                names.append(line.split()[1])\n",
    "                section_insides.append([])\n",
    "            section_insides[-1].append(line)\n",
    "        return section_insides, names\n",
    "    \n",
    "    def get_section_content(self, section_name, order=0):\n",
    "        # znalezc wszystkie instancje nazwy \"section_name\" w self.sections_names\n",
    "        # ustalic, o ktora chodzi (order) -> numer sekcji w self.sections\n",
    "        # zwrocic odpowiedni element self.sections\n",
    "        n = len(self.sections_names)\n",
    "        indegz = [j for j in range(n) if self.sections_names[j]==section_name][order]\n",
    "        return self.sections[indegz]\n",
    "     \n",
    "    def set_section_content(self, section_name, order, line_num, new_line):\n",
    "        n = len(self.sections_names)\n",
    "        indegz = [j for j in range(n) if self.sections_names[j]==section_name][order]\n",
    "        self.sections[indegz][line_num] = new_line\n",
    "    \n",
    "    def add_state_B(self):\n",
    "        section_content = self.get_section_content('atoms')\n",
    "        # w petli po liniach:\n",
    "        # (1) sprawdzic, czy linia jest legitna\n",
    "        # (2) znalezc typ, ladunek i mase\n",
    "        # (3) dodac w odp kolejnosci do linii\n",
    "        # (4) zapisac zmodyfikowana linie w odpowiednim elemencie self.sections\n",
    "        num = 0\n",
    "        for linenum, line in enumerate(section_content):\n",
    "            if len(line.split()) > 7 and not line.lstrip().startswith(';') and str(line.split()[2])==str(self.resnum):\n",
    "                nazwa, typ, masa = line.split()[4], line.split()[1], line.split()[7]\n",
    "                try:\n",
    "                    ladunek = self.charges[nazwa] \n",
    "                except KeyError:\n",
    "                    ladunek = \"0.00\"\n",
    "                if nazwa in self.names_to_types_B.keys():\n",
    "                    typ = self.names_to_types_B[nazwa]\n",
    "                else:\n",
    "                    typ = 'DH'\n",
    "                    self.names_to_types_B[nazwa] = 'DH'\n",
    "                line = line.rstrip()\n",
    "                if ';' in line:\n",
    "                    line = line[:line.index(';')]\n",
    "                suffix = \"{:10s} {:10s} {:10s}\".format(typ, ladunek, masa)\n",
    "                new_line = line + suffix + '\\n'\n",
    "                self.set_section_content('atoms', 0, linenum, new_line)\n",
    "                num += 1   \n",
    "    \n",
    "    def all_sections_add_B(self):\n",
    "        self.add_B_to_section('bonds', 2)\n",
    "        self.add_B_to_section('angles', 3)\n",
    "        self.add_B_to_section('dihedrals', 4)\n",
    "    \n",
    "    def add_B_to_section(self, section, n_entries):\n",
    "        section_content = self.get_section_content(section)\n",
    "        for linenum, line in enumerate(section_content):\n",
    "            atomnum_list = line.split()[0:n_entries]\n",
    "            if any([True for i in atomnum_list if i in self.different_types]):\n",
    "                parA=[self.names_to_types_A[self.nums_to_names[l]] for l in atomnum_list]\n",
    "                if parA is None:\n",
    "                    raise RuntimeError('{} not found in ffbonded'.format(' '.join(parA)))\n",
    "                parB=[self.names_to_types_B[self.nums_to_names[l]] for l in atomnum_list]\n",
    "                form=\"{:10} \"*len(self.find_param(parA))*2\n",
    "                try:\n",
    "                    suffix = form.format(*self.find_param(parA), *self.find_param(parB))\n",
    "                except KeyError:\n",
    "                    suffix = form.format(*self.find_param(parA), *self.find_param(parA))\n",
    "                new_line = line.rstrip() + suffix + '\\n'\n",
    "                self.set_section_content(section,0,linenum,new_line)\n",
    "        \n",
    "    def find_param(self, atomtypes_list):\n",
    "        nparams = len(atomtypes_list)\n",
    "        interaction_types = ['1', '2', '4', '5', '9']\n",
    "        param = None\n",
    "        for line in self.ffbonded[::-1]:\n",
    "            lspl = line.split()\n",
    "            if len(lspl) > nparams and lspl[nparams] in interaction_types:\n",
    "                if self.test_line(lspl[:nparams], atomtypes_list):\n",
    "                    if ';' not in lspl[nparams+1:]:\n",
    "                        param = [float(x) for x in lspl[nparams+1:]]\n",
    "                    else:\n",
    "                        endpos = lspl.index(';')\n",
    "                        param = [float(x) for x in lspl[nparams+1:endpos]]\n",
    "        if param:\n",
    "            return param\n",
    "        else:\n",
    "            raise KeyError()\n",
    "    \n",
    "    def test_line(self, line_list, ref_list):\n",
    "        assert len(line_list) == len(ref_list)\n",
    "        if 'X' in line_list:\n",
    "            non_x_nums = [x for x in range(len(line_list)) if line_list[x] != \"X\"]\n",
    "            line_list = [line_list[a] for a in non_x_nums]\n",
    "            ref_list = [ref_list[a] for a in non_x_nums]\n",
    "        n = len(line_list)\n",
    "        if all([line_list[i] == ref_list[i] for i in range(n)]) or all([line_list[i] == ref_list[n-i-1] for i in range(n)]):\n",
    "            return True\n",
    "        else:\n",
    "            return False\n",
    "    \n",
    "    def write(self):\n",
    "        with open(self.prefix + self.filename, 'w') as outfile:\n",
    "            for section in self.sections:\n",
    "                for line in section:\n",
    "                    outfile.write(line)\n",
    "                    \n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "q = Alchemist('topol.itp', 23, 'GLU', '/home/miloszw/gmx5_plu21/share/gromacs/top/amber99bsc1.ff/ffbonded.itp', '/home/miloszw/gmx5_plu21/share/gromacs/top/amber99bsc1.ff/aminoacids.rtp')\n",
    "\n",
    "q()\n"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (Anaconda 2019)",
   "env": {
    "AR": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-ar",
    "AS": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-as",
    "CC": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-cc",
    "CONDA_EXE": "/ext/anaconda-2019.03/bin/conda",
    "CONDA_PREFIX": "/ext/anaconda-2019.03",
    "CONDA_PYTHON_EXE": "/ext/anaconda-2019.03/bin/python",
    "CPP": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-cpp",
    "CXX": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-c++",
    "CXXFILT": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-c++filt",
    "ELFEDIT": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-elfedit",
    "F77": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gfortran",
    "F90": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gfortran",
    "F95": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-f95",
    "FC": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gfortran",
    "GCC": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gcc",
    "GCC_AR": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gcc-ar",
    "GCC_NM": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gcc-nm",
    "GCC_RANLIB": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gcc-ranlib",
    "GDAL_DATA": "/ext/anaconda-2019.03/share/gdal",
    "GFORTRAN": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gfortran",
    "GPROF": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-gprof",
    "GXX": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-g++",
    "JAVA_HOME": "/ext/anaconda-2019.03",
    "JAVA_LD_LIBRARY_PATH": "/ext/anaconda-2019.03/lib/server",
    "LD": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-ld",
    "LD_GOLD": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-ld.gold",
    "LD_LIBRARY_PATH": "/ext/anaconda-2019.03/lib",
    "NM": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-nm",
    "OBJCOPY": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-objcopy",
    "OBJDUMP": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-objdump",
    "OCAMLFIND_CONF": "/ext/anaconda-2019.03/etc/findlib.conf",
    "OCAMLLIB": "/ext/anaconda-2019.03/lib/ocaml",
    "OCAML_PREFIX": "/ext/anaconda-2019.03",
    "PATH": "/ext/anaconda-2019.03/bin:/ext/anaconda-2019.03/condabin:/cocalc/bin:/cocalc/src/smc-project/bin:/home/user/bin:/home/user/.local/bin:/ext/bin:/usr/lib/xpra:/opt/ghc/bin:/usr/local/bin:/usr/bin:/bin:/ext/data/homer/bin:/ext/data/weblogo:/ext/intellij/idea/bin:/ext/pycharm/pycharm/bin:/usr/lib/postgresql/10/bin",
    "PROJ_LIB": "/ext/anaconda-2019.03/share/proj",
    "PYTHONHOME": "/ext/anaconda-2019.03/lib/python3.7",
    "PYTHONPATH": "/ext/anaconda-2019.03/lib/python3.7:/ext/anaconda-2019.03/lib/python3.7/site-packages",
    "RANLIB": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-ranlib",
    "READELF": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-readelf",
    "RSTUDIO_WHICH_R": "/ext/anaconda-2019.03/bin/R",
    "SIZE": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-size",
    "STRINGS": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-strings",
    "STRIP": "/ext/anaconda-2019.03/bin/x86_64-conda_cos6-linux-gnu-strip"
   },
   "language": "python",
   "metadata": {
    "cocalc": {
     "description": "Python/R distribution for data science",
     "priority": 5,
     "url": "https://www.anaconda.com/distribution/"
    }
   },
   "name": "anaconda2019",
   "resource_dir": "/ext/jupyter/kernels/anaconda2019"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.5.5"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}